#! /usr/bin/env python
"""
@author: Ajay Arunachalam
Created on: 13/10/2021
Goal: Filtered and Original Raw Time-series data Similarity with Siamese Neural Network using keras
Version: 0.0.5
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from keras.models import Model
from keras.layers import Input, Subtract, Multiply, Concatenate, Dense, BatchNormalization, Dropout, Activation
from keras import backend as K
import warnings
warnings.filterwarnings('ignore')

class Siamese:

	def Siamese_Model(features):

	    """
	    Implementation of the Siamese Network
	    Args:
	        features (int): number of features 
	    Returns:
	        [keras model]: siamese model
	    """

	    inp1 = Input(shape=(features,))
	    inp2 = Input(shape=(features,))

	    diff = Subtract()([inp1,inp2])
	    # squared difference
	    L2 = Multiply()([diff,diff])                     
	    # product proximity
	    prod = Multiply()([inp1,inp2])                  
	    # combined metric
	    combine = Concatenate(axis=1)([L2,prod])                

	    path1 = Dense(64)(L2)
	    path1 = BatchNormalization()(path1)
	    path1 = Dropout(0.25)(path1)
	    path1 = Activation('relu')(path1)

	    path2 = Dense(64)(prod)
	    path2 = BatchNormalization()(path2)
	    path2 = Dropout(0.25)(path2)
	    path2 = Activation('relu')(path2)

	    path3 = Dense(64)(combine)
	    path3 = BatchNormalization()(path3)
	    path3 = Dropout(0.25)(path3)
	    path3 = Activation('relu')(path3)

	    # combining everything
	    paths = Concatenate(axis=1)([path1,path2,path3])        


	    top = Dense(256)(paths)
	    top = BatchNormalization()(top)
	    top = Dropout(0.25)(top)
	    top = Activation('relu')(top)


	    out = Dense(1)(top)   # output similarity score                          

	    siamese_model = Model(inputs=[inp1,inp2],outputs=[out])

	    return siamese_model


#! /usr/bin/env python

"""
@author: Ajay Arunachalam
Created on: 06/10/2021
Helper and Utility functions for forecasting model pipeline
Version: 0.0.5
"""
from math import sqrt
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler, StandardScaler, MaxAbsScaler, RobustScaler
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score, explained_variance_score
from sklearn.linear_model import LinearRegression
import plotly.offline as pyo
import plotly.graph_objs as go
from plotly.offline import iplot
import torch
import os
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from .forecast_ml import *
from .forecast_ml_extension import *
from .stats import *
import seaborn as sns
from random import *
get_ipython().run_line_magic('matplotlib', 'inline')

class Helper:

	def get_variable(df, timestamp_col, forecasting_col):
		orig_df = df.copy(deep=True)
		df = df.set_index([timestamp_col])
		df = df.rename(columns={forecasting_col: 'value'})

		df.index = pd.to_datetime(df.index)
		if not df.index.is_monotonic:
			df = df.sort_index()
		return df, orig_df


	def predictor_outcome_split(df, target_col):
		#y = df.iloc[:,target_col]
		#X = df.iloc[:,target_col+1:df.shape[1]]
		y = df[[target_col]]
		X = df.drop(columns=[target_col])
		return X, y

	def train_val_test_split(df, target_col, test_ratio):
		val_ratio = test_ratio / (1 - test_ratio)
		#val_ratio = 0.25 #test_ratio / (1 - test_ratio)
		X, y = Helper.predictor_outcome_split(df, target_col)
		X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_ratio, shuffle=False)
		X_train, X_val, y_train, y_val = train_test_split(X_train, y_train, test_size=val_ratio, shuffle=False)
		print("Train-Val-Test Split")
		print(f'Predictors: Train-{(X_train.shape)}, Val-{(X_val.shape)}, Test-{(X_test.shape)}')
		print(f'Response: Train-{(y_train.shape)}, Val-{(y_val.shape)}, Test-{(y_test.shape)}')
		return X_train, X_val, X_test, y_train, y_val, y_test

	def get_scaler(scaler):
		scalers = {
		"minmax": MinMaxScaler,
		"standard": StandardScaler,
		"maxabs": MaxAbsScaler,
		"robust": RobustScaler,
		}
		return scalers.get(scaler.lower())()

	def apply_transformation(X_train, X_val, X_test, y_train, y_val, y_test, scaler):

		def get_scaler(scaler):
			scalers = {
				"minmax": MinMaxScaler,
				"standard": StandardScaler,
				"maxabs": MaxAbsScaler,
				"robust": RobustScaler,
			}
			return scalers.get(scaler.lower())()

		scaler = get_scaler(scaler)
		X_train_arr = scaler.fit_transform(X_train)
		X_val_arr = scaler.transform(X_val)
		X_test_arr = scaler.transform(X_test)

		y_train_arr = scaler.fit_transform(y_train)
		y_val_arr = scaler.transform(y_val)
		y_test_arr = scaler.transform(y_test)

		#X_full_arr = np.concatenate(X_train_arr, X_val_arr, X_test_arr)
		#y_full_arr = np.concatenate(y_train_arr, y_val_arr, y_test_arr)
		return X_train_arr, X_val_arr, X_test_arr, y_train_arr, y_val_arr, y_test_arr, scaler

	def apply_transformation_forecast(X, scaler):
		scaler = Helper.get_scaler(scaler)
		X_arr = scaler.fit_transform(X)
		return X_arr

	def prepare_pytorch_data_forecast_df(X_arr):
		unseen_loader = torch.Tensor(X_arr)
		print(f'Tensor size: {unseen_loader.size(0)}')
		return unseen_loader

	def prepare_pytorch_data(X_train_arr, X_val_arr, X_test_arr, y_train_arr, y_val_arr, y_test_arr, batch_size):
		import torch
		from torch.utils.data import TensorDataset, DataLoader
		train_features = torch.Tensor(X_train_arr)
		train_targets = torch.Tensor(y_train_arr)
		val_features = torch.Tensor(X_val_arr)
		val_targets = torch.Tensor(y_val_arr)
		test_features = torch.Tensor(X_test_arr)
		test_targets = torch.Tensor(y_test_arr)
		
		train = TensorDataset(train_features, train_targets)
		val = TensorDataset(val_features, val_targets)
		test = TensorDataset(test_features, test_targets)
		
		train_loader = DataLoader(train, batch_size=batch_size, shuffle=False, drop_last=True)
		val_loader = DataLoader(val, batch_size=batch_size, shuffle=False, drop_last=True)
		test_loader = DataLoader(test, batch_size=batch_size, shuffle=False, drop_last=True)
		test_loader_one = DataLoader(test, batch_size=1, shuffle=False, drop_last=True)
		
		return train_loader, val_loader, test_loader, test_loader_one

	def get_stats_model(model):
		models = {
			"em": Stats_Models.EMModel,

		}
		return models.get(model.lower())

	def get_model(model, model_params):
		models = {
			"rnn": Forecasting_Models.RNNModel,
			"lstm": Forecasting_Models.LSTMModel,
			"gru": Forecasting_Models.GRUModel,
			"birnn": Forecasting_Models.BiRNNModel,
			"bigru": Forecasting_Models.BiGRUModel,
			"cnn": Forecasting_Models.CNNModel,
			"deepcnn": Forecasting_Models.DeepCNNModel,
			
		}
		return models.get(model.lower())(**model_params)

	def inverse_transform(scaler, df, columns):
		for col in columns:
			df[col] = scaler.inverse_transform(df[col])
		return df


	def format_predictions(predictions, values, df_test, scaler):
		vals = np.concatenate(values, axis=0).ravel()
		preds = np.concatenate(predictions, axis=0).ravel()
		df_result = pd.DataFrame(data={"value": vals, "prediction": preds}, index=df_test.head(len(vals)).index)
		df_result = df_result.sort_index()
		df_result = Helper.inverse_transform(scaler, df_result, [["value", "prediction"]])
		return df_result

	def forecast_bias():
		pass

	def smape(actual, prediction):
		dividend= np.abs(np.array(actual) - np.array(prediction))
		denominator = np.array(actual) + np.array(prediction)
		return 2 * np.mean(np.divide(dividend, denominator, out=np.zeros_like(dividend), where=denominator!=0, casting='unsafe'))

	def calculate_metrics(df):
		metrics_one = {'mae' : mean_absolute_error(df.value, df.prediction),
						  'rmse' : mean_squared_error(df.value, df.prediction) ** 0.5,
						   'fc_bias' : (np.sum(np.array(df.value) - np.array(df.prediction)) * 1.0/len(df.value)),
						   'mape' : np.mean((np.abs((df.value-df.prediction) / df.value)) * 100) ,
					}
		
		print("Mean Absolute Error (MAE):       ", metrics_one["mae"])
		print("Root Mean Squared Error (RMSE):   ", metrics_one["rmse"])
		print("Forecast bias:   		  ", metrics_one["fc_bias"])
		print("Mean Absolute Percentage Error (MAPE):   ", metrics_one["mape"])
	

		metrics_two = {'r2' : r2_score(df.value, df.prediction),
						  'evs': explained_variance_score(df.value, df.prediction),
						  'rmsre' : np.sqrt(np.mean(((df.value-df.prediction)/df.value)**2)),
						  'smape': Helper.smape(df.value, df.prediction),
						  'r': np.corrcoef(df.value, df.prediction)[0, 1],}

		print("R^2 Score:                 ", metrics_two["r2"])
		print("Explained Variance Score:  ", metrics_two["evs"])
		print("Root Mean Squared Relative Error (RMSRE): ", metrics_two["rmsre"])
		print("Symmetric Mean Absolute Percentage Error (sMAPE):   ", metrics_two["smape"])
		print("pcc coefficient: ", metrics_two["r"])

		return metrics_one, metrics_two

	def plot_metrics(metrics_one, metrics_two):
		# Visualize Metrics as bar plot with sns
		data_metrics_one = pd.DataFrame.from_dict(metrics_one.items())

		fig, axes = plt.subplots(nrows=1, ncols=1, figsize=(10,7))
		axes = sns.barplot(data = data_metrics_one, x = 0, y=1, hue=0)
		axes.set(xlabel = 'Metrics', ylabel='Value', title='Model Peformance wrt. their corresponding metrics')
		fig.savefig('./metric_plot_1.png')

		data_metrics_two = pd.DataFrame.from_dict(metrics_two.items())

		fig, axes = plt.subplots(nrows=1, ncols=1, figsize=(10,7))
		axes = sns.barplot(data = data_metrics_two, x = 0, y=1, hue=0)
		axes.set(xlabel = 'Metrics', ylabel='Value', title='Model Peformance wrt. their corresponding metrics')
		fig.savefig('./metric_plot_2.png')

	# def make_future_df(df, ts, period:int, fq:str):
	# 	last_date = df[ts].max()
	# 	dates = pd.date_range(start=last_date, periods=period + 1, freq=fq)
	# 	dates = dates[dates > last_date]  # Drop start if equals last_date
	# 	dates = dates[:period]  # Return correct number of periods 
	# 	return pd.Dataframe({ts:dates})

	def make_future_df(df, ts, period:int, fq:str):
		#last_date = df.index.max()
		#dates = pd.date_range(start=last_date, periods=period + 1, freq=fq)
		#assert dates == dates
		digit = ''.join(filter(str.isdigit, str(fq)))
		interval = int(digit)
		if str(fq) == 'h' or str(fq) == 'H' or interval is not None:  # hourly
			#last_date = df[ts].max()
			last_date = df.index.max()
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[:period]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		elif str(fq) == 'm' or str(fq) == 'M' or interval is not None: # monthly
			last_date = df.index.max()
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[1:period+1]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		elif str(fq) == 'd' or str(fq) == 'D' or interval is not None: # daily
			last_date = df.index.max()
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[1:period+1]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		elif str(fq) == 'w' or str(fq) == 'W' or interval is not None: # weekly
			last_date = df.index.max()
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[1:period+1]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		elif str(fq) == 'min' or str(fq) == 'MIN' or interval is not None: # minute
			last_date = df.index.max()
			diff_min = (df.index.minute[1] - df.index.minute[0])
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[:period]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		elif str(fq) == 's' or str(fq) == 'S' or interval is not None: # second
			last_date = df.index.max()
			diff_sec = (df.index.second[1] - df.index.second[0])
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[:period]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		elif str(fq) == 'n' or str(fq) == 'N' or interval is not None: # millisecond
			last_date = df.index.max()
			diff_sec = (df.index.second[1] - df.index.second[0])
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[:period]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		elif str(fq) == 'q' or str(fq) == 'Q' or str(fq) == 'QS': # quaterly
			last_date = df.index.max()
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[:period]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		elif str(fq) == '2q' or str(fq) == '2Q' or str(fq) == 'HA': # half yearly
			last_date = df.index.max()
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[:period]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		elif str(fq) == 'y' or str(fq) == 'Y' or str(fq) == 'A' or interval is not None: # yearly
			last_date = df.index.max()
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[:period]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		return ff

	def make_future_dataframe(self, periods, freq='D', include_history=True):
		
		"""Simulate the trend using the extrapolated generative model.
		Parameters
		----------
		periods: Int number of periods to forecast forward.
		freq: Any valid frequency for pd.date_range, such as 'D' or 'M'.
		include_history: Boolean to include the historical dates in the data
			frame for predictions.
		Returns
		-------
		pd.Dataframe that extends forward from the end of self.history for the
		requested number of periods.
		"""
		if self.history_dates is None:
			raise Exception('Model has not been fit.')
		last_date = self.history_dates.max()
		dates = pd.date_range(
			start=last_date,
			periods=periods + 1,  # An extra in case we include start
			freq=freq)
		dates = dates[dates > last_date]  # Drop start if equals last_date
		dates = dates[:periods]  # Return correct number of periods

		if include_history:
			dates = np.concatenate((np.array(self.history_dates), dates))

		return pd.DataFrame({'ds': dates})


	# def forecast_window_inference(predictions, df, scaler):
	# 	preds = np.concatenate(predictions, axis=0).ravel()
	# 	df_result = pd.DataFrame(data={"prediction": preds}, index=df_test.head(len(df)).index)
	# 	df_result = df_result.sort_index()
	# 	df_result = Helper.inverse_transform(scaler, df_result, [["prediction"]])
	# 	return df_result

	def forecast_window_inference(predictions, df, scaler):
		preds = np.concatenate(predictions, axis=0).ravel()
		df_result = pd.DataFrame(data={"value": preds}, index=df.head(len(preds)).index)
		df_result = df_result.sort_index()
		df_result = Helper.inverse_transform(scaler, df_result, [["value"]])
		return df_result

	def save_final_data(df, ff, ts, fc):
		forecasted_data = pd.concat([df.reset_index(), ff.reset_index()], axis=0)
		forecasted_data.set_index(ts, inplace=True)
		forecasted_data = forecasted_data.rename(columns={'value':fc})
		fig = forecasted_data.iloc[df.shape[0]:].plot().get_figure()
		plt.tight_layout()
		fig.savefig("./model_forecast.png", dpi=150, bbox_inches='tight')
		forecasted_data.to_csv('./model_full_data.csv', encoding='utf-8')
		return forecasted_data

	
	def compare_ml_models_and_plot():
		# Install regressormetricgraphplot package from terminal or notebook
		'''
		Terminal: 
		$ pip install regressormetricgraphplot
		 OR
		$ git clone https://github.com/ajayarunachalam/RegressorMetricGraphPlot
		$ cd RegressorMetricGraphPlot
		$ python setup.py install
		Notebook:
		!git clone https://github.com/ajayarunachalam/RegressorMetricGraphPlot.git
		cd RegressorMetricGraphPlot/
		Just replace the line 'from CompareModels import *' with 'from regressormetricgraphplot import CompareModels' 
		'''
		# Now, let us check how machine learning algorithms perform on this dataset in comparison to the build neural network
		from regressioncomparemetricplot import CompareModels

		# Linear Regression 

		# Fitting training set to linear regression model
		lr = LinearRegression(n_jobs=-1)
		lr.fit(X_train, y_train)

		# Predicting the house price
		y_pred = lr.predict(X_test)

		# Metrics
		print(f'R2_nd_RMSE LR MODEL: {CompareModels.R2AndRMSE(y_test=y_test, y_pred=y_pred)}')

		plot = CompareModels()

		plot.add(model_name='Linear Regression', y_test=y_test, y_pred=y_pred)

		plot.show(figsize=(10, 5))

		# Fitting Random Forest model to the dataset
		rfr = RandomForestRegressor(n_estimators=10, random_state=10, n_jobs=-1)
		rfr.fit(X_train, y_train)

		# Predicting the outcome
		y_pred = rfr.predict(X_test)

		print(f'R2_nd_RMSE RF MODEL: {CompareModels.R2AndRMSE(y_test=y_test, y_pred=y_pred)}')

		plot.add('Random Forest', y_test, y_pred)

		plot.show(figsize=(10, 5))

		xgb = XGBRegressor(n_jobs=4, silent=False, objective='reg:linear',
					   max_depth=3, random_state=10, n_estimators=100,
					   learning_rate=0.3, verbose=True)

		xgb.fit(X_train, y_train)

		# Predicting the outcome
		y_pred = xgb.predict(X_test)

		print(f'R2_nd_RMSE XGB MODEL: {CompareModels.R2AndRMSE(y_test=y_test, y_pred=y_pred)}')

		plot.add('XGBoost', y_test, y_pred)

		plot.show(figsize=(10, 5))

	def build_baseline_model(df, test_ratio, target_col):
		X, y = Helper.predictor_outcome_split(df, target_col)
		X_train, X_test, y_train, y_test = train_test_split(
			X, y, test_size=test_ratio, shuffle=False
		)
		model = LinearRegression()
		model.fit(X_train, y_train)
		prediction = model.predict(X_test)

		result = pd.DataFrame(y_test)
		result["prediction"] = prediction
		result = result.sort_index()

		return result

	def plot_predictions(df_result, df_baseline):
		# Create the plotly figure
		data = []
		value = go.Scatter(
			x=df_result.index,
			y=df_result.value,
			mode="lines",
			name="actual values",
			marker=dict(),
			text=df_result.index,
			line=dict(color="rgba(0,0,0, 0.3)"),
		)
		data.append(value)

		baseline = go.Scatter(
			x=df_baseline.index,
			y=df_baseline.prediction,
			mode="lines",
			line={"dash": "dot"},
			name='linear regression',
			marker=dict(),
			text=df_baseline.index,
			opacity=0.8,
		)
		data.append(baseline)
		
		prediction = go.Scatter(
			x=df_result.index,
			y=df_result.prediction,
			mode="lines",
			line={"dash": "dot"},
			name='deep model predictions',
			marker=dict(),
			text=df_result.index,
			opacity=0.8,
		)
		data.append(prediction)
		
		layout = dict(
			title="Predictions vs Actual Values for the dataset",
			xaxis=dict(title="Time", ticklen=5, zeroline=False),
			yaxis=dict(title="Value", ticklen=5, zeroline=False),
		)

		fig = dict(data=data, layout=layout)
		iplot(fig)

	# plot only user provided time-window
	def plot_forecast(ff_result, fc):
		# Create the plotly figure
		data = []
		forecast = go.Scatter(
			x=ff_result.index,
			y=ff_result.value,
			mode="lines",
			line={"dash": "dot"},
			name='Forecasted Results',
			marker=dict(),
			text=ff_result.index,
			opacity=0.8,
			
		)
		data.append(forecast)
		
		layout = dict(
			title="Forecast for the user provided time period",
			xaxis=dict(title=f"{ff_result.index.name}", ticklen=5, zeroline=False),
			yaxis=dict(title=f"{fc}", ticklen=5, zeroline=False),
		)

		fig = dict(data=data, layout=layout)
		iplot(fig)

	def plot_forecast_complete(ff, fc, FORECAST_PERIOD):
		# Create the plotly figure
		index = ff.columns.get_loc(fc)
		actual = go.Scatter(
		  x=ff.index[:-FORECAST_PERIOD],
		  y=ff.iloc[:-FORECAST_PERIOD,index],
		  mode = 'markers',
		  marker = {
			'color': '#FF0000',
			'size': 1,
			},
			line= {
			  'width': 2
			},
		  name = 'Actual'
		)

		forecast = go.Scatter(
		x=ff.index[-FORECAST_PERIOD:],
		  y=ff.iloc[-FORECAST_PERIOD:, index],
		  mode = 'markers',
		  marker = {
		'color': '#3bbed7',
		'size': 1,
		  },
		  line = {
			'width': 2
		  },
		  name = 'Forecast',
		)

		data = [actual, forecast]

		layout = dict(
			title="Forecast vs Historical Values for the dataset",
				xaxis=dict(title=f"{ff.index.name}", ticklen=5, zeroline=False),
				yaxis=dict(title=f'{fc}', ticklen=5, zeroline=False),
		)

		fig = dict(data=data, layout=layout)
		iplot(fig)

	def plot_nowcast(ff, fc, FORECAST_PERIOD):
		# Create the plotly figure
		actual = go.Scatter(
		  x=ff.index[:-FORECAST_PERIOD],
		  y=ff.loc[:-FORECAST_PERIOD,fc],
		  mode = 'markers',
		  marker = {
			'color': '#fffaef',
			'size': 3,
			'line': {
			  'color': '#000000',
			  'width': .75
			}
		  },
		  name = 'Actual'
		)

		nowcast = go.Scatter(
		x=ff.index[-FORECAST_PERIOD:],
		  y=ff.loc[-FORECAST_PERIOD:,fc],
		  mode = 'lines',
		  marker = {
		'color': '#3bbed7'
		  },
		  line = {
			'width': 3
		  },
		  name = 'Nowcast',
		)

		data = [actual, nowcast]

		layout = dict(
			title="Nowcast vs Historical Values for the dataset",
				xaxis=dict(title=f"{ff.index.name}", ticklen=5, zeroline=False),
				yaxis=dict(title=f"{fc}", ticklen=5, zeroline=False),
		)

		fig = dict(data=data, layout=layout)
		iplot(fig)



	def explainable_forecast(df, ff, fc, specific_prediction_sample_to_explain:int, input_label_index_value, num_labels:int):

		"""
		Understand, interpret, and trust the results of the deep learning models at individual/samples level and multiple columns
		"""
		import shap
		import numpy as np
		import pandas as pd
		from keras.models import Sequential
		from keras.layers import Dense
		import ipywidgets as widgets

		data = pd.concat([df, ff], axis=0)
		data.rename(columns={"value":fc}, inplace=True)

		X, Y = Helper.predictor_outcome_split(data, fc)

		# Get the number of inputs and outputs from the dataset
		n_inputs, n_outputs = X.shape[1], Y.shape[1]

		def get_model(n_inputs, n_outputs):
			model_nn = Sequential()
			model_nn.add(Dense(32, input_dim=n_inputs, kernel_initializer='he_uniform', activation='relu'))
			model_nn.add(Dense(n_outputs, kernel_initializer='he_uniform'))
			model_nn.compile(loss='mae', optimizer='adam')
			return model_nn

		model_nn = get_model(n_inputs, n_outputs)

		# model_nn.fit(X.iloc[10:,:].values, Y, epochs=30)

		# model_nn.evaluate(x = X.iloc[10:,:].values, y = Y)

		model_nn.fit(X.iloc[:df.shape[0],:].values, Y.iloc[:df.shape[0]], epochs=30)

		model_nn.evaluate(x = X.iloc[:df.shape[0],:].values, y = Y.iloc[:df.shape[0]])

		XpredictInputData = X.iloc[specific_prediction_sample_to_explain,:] # X[specific_prediction_sample_to_explain,:]

		if (XpredictInputData.ndim == 1):
			XpredictInputData = np.array([XpredictInputData])

		print(model_nn.predict(XpredictInputData)) # 0:1

		'''
		Here we take the Keras model trained above and explain why it makes different predictions on individual samples.

		Set the explainer using the Kernel Explainer (Model agnostic explainer method form SHAP).
		'''
		explainer = shap.KernelExplainer(model = model_nn.predict, data = X.head(50), link = "identity") # data = X[0:50]

		'''
		Get the Shapley value for a single example.
		'''
		# Set the index of the specific example to explain

		shap_value_single = explainer.shap_values(X = X.iloc[specific_prediction_sample_to_explain,:], nsamples = 100)  # X[specific_prediction_sample_to_explain,:]

		'''
		Display the details of the single example
		'''
		print(X.iloc[specific_prediction_sample_to_explain,:]) 
		'''
		Choose the label/output/target to run individual explanations on:

		Note: The dropdown menu can easily be replaced by manually setting the index on the label to explain.
		'''
		# Create the list of all labels for the drop down list
		#label_cols = ['window_diff_0', 'window_diff_1', 'window_diff_2', 'window_diff_3', 'window_diff_4', 'window_diff_5', 'window_diff_6']

		#label_cols = ['window_diff_'+str(i) for i in range(num_labels)]
		label_cols = [f'{fc}_'+str(i) for i in range(num_labels)]
		#print(label_cols)
		df_labels = pd.DataFrame(data = Y, columns = label_cols)
		#df_labels.to_csv('./y_labels.csv')
		list_of_labels = df_labels.columns.to_list() # Y.columns.to_list()

		# Create a list of tuples so that the index of the label is what is returned
		tuple_of_labels = list(zip(list_of_labels, range(len(list_of_labels))))

		# Create a widget for the labels and then display the widget
		current_label = widgets.Dropdown(options=tuple_of_labels,
									  value=input_label_index_value,
									  description='Select Label:'
									  )

		# Display the dropdown list (Note: access index value with 'current_label.value')
		print(current_label)
		#Dropdown(description='Select Label:', options=(('labels_01', 0), ('labels_02', 1), ('labels_03', 2), etc

		'''
		Plot the force plot for a single example and a single label/output/target
		'''
		print(f'Current label Shown: {list_of_labels[current_label.value]}')

		# print the JS visualization code to the notebook
		shap.initjs()

		shap.force_plot(base_value = explainer.expected_value[current_label.value],
						shap_values = shap_value_single[current_label.value], 
						features = X.iloc[specific_prediction_sample_to_explain,:] # X_idx:X_idx+1
						)

		'''
		Create the summary plot for a specific output/label/target.
		'''
		# Note: We are limiting to the first 50 training examples since it takes time to calculate the full number of sampels
		shap_values = explainer.shap_values(X = X.iloc[0:50,:], nsamples = 100) # X[0:50,:]

		print(f'Current Label Shown: {list_of_labels[current_label.value]}\n')

		# print the JS visualization code to the notebook
		shap.initjs()

		shap.summary_plot(shap_values = shap_values[current_label.value],
				  features = X.iloc[0:50,:],
				  plot_type="bar", # X[0:50,:]
				  show=False
				  )

		plt.savefig('./forecast_model_summary_plot.png', dpi=150, bbox_inches='tight')

		'''
		Force Plot for the first 50 individual examples.
		'''
		print(f'Current Label Shown: {list_of_labels[current_label.value]}\n')

		# print the JS visualization code to the notebook
		shap.initjs()

		shap.force_plot(base_value = explainer.expected_value[current_label.value],
						shap_values = shap_values[current_label.value], 
						features = X.iloc[0:50,:] # X[0:50,:]
						)
		

class Optimization:

	"""Optimization is a helper class that allows training, validation, prediction.

	Optimization is a helper class that takes model, loss function, optimizer function
	learning scheduler (optional), early stopping (optional) as inputs. In return, it
	provides a framework to train and validate the models, and to predict future values
	based on the models.

	Attributes:
		model (RNNModel, LSTMModel, GRUModel): Model class created for the type of RNN
		loss_fn (torch.nn.modules.Loss): Loss function to calculate the losses
		optimizer (torch.optim.Optimizer): Optimizer function to optimize the loss function
		train_losses (list[float]): The loss values from the training
		val_losses (list[float]): The loss values from the validation
		last_epoch (int): The number of epochs that the models is trained
	"""
	def __init__(self, device, model, loss_fn, optimizer):

		"""
		Args:
			model (RNNModel, LSTMModel, GRUModel): Model class created for the type of RNN
			loss_fn (torch.nn.modules.Loss): Loss function to calculate the losses
			optimizer (torch.optim.Optimizer): Optimizer function to optimize the loss function
		"""
		self.device = device
		self.model = model
		self.loss_fn = loss_fn
		self.optimizer = optimizer
		self.train_losses = []
		self.val_losses = []
		

	
	def train_step(self, x, y):
		"""The method train_step completes one step of training.

		Given the features (x) and the target values (y) tensors, the method completes
		one step of the training. First, it activates the train mode to enable back prop.
		After generating predicted values (yhat) by doing forward propagation, it calculates
		the losses by using the loss function. Then, it computes the gradients by doing
		back propagation and updates the weights by calling step() function.

		Args:
			x (torch.Tensor): Tensor for features to train one step
			y (torch.Tensor): Tensor for target values to calculate losses

		"""
		# Sets model to train mode
		self.model.train().to(self.device)

		# Makes predictions
		yhat = self.model(x).to(self.device)

		# Computes loss
		loss = self.loss_fn(y, yhat)

		# Computes gradients
		loss.backward()

		# Updates parameters and zeroes gradients
		self.optimizer.step()
		self.optimizer.zero_grad()

		# Returns the loss
		return loss.item()

	def train(self, train_loader, val_loader, batch_size, n_epochs, n_features):
		"""The method train performs the model training

		The method takes DataLoaders for training and validation datasets, batch size for
		mini-batch training, number of epochs to train, and number of features as inputs.
		Then, it carries out the training by iteratively calling the method train_step for
		n_epochs times. If early stopping is enabled, then it  checks the stopping condition
		to decide whether the training needs to halt before n_epochs steps. Finally, it saves
		the model in a designated file path.

		Args:
			train_loader (torch.utils.data.DataLoader): DataLoader that stores training data
			val_loader (torch.utils.data.DataLoader): DataLoader that stores validation data
			batch_size (int): Batch size for mini-batch training
			n_epochs (int): Number of epochs, i.e., train steps, to train
			n_features (int): Number of feature columns

		"""
		os.makedirs('./model_path/', exist_ok=True)
		path = "./model_path/"
		#os.path.join(path,“filename.pth”)
		#model_path = f'{path}deployed_model/{self.model}_{datetime.now().strftime("%Y-%m-%d %H:%M:%S")}'
		#model_path = f'{path}deployed_model/_{datetime.now().strftime("%Y-%m-%d_%H:%M:%S")}'

		for epoch in range(1, n_epochs + 1):
			batch_losses = []
			for x_batch, y_batch in train_loader:
				x_batch = x_batch.view([batch_size, -1, n_features]).to(self.device)
				y_batch = y_batch.to(self.device)
				loss = self.train_step(x_batch, y_batch)
				batch_losses.append(loss)
			training_loss = np.mean(batch_losses)
			self.train_losses.append(training_loss)

			with torch.no_grad():
				batch_val_losses = []
				for x_val, y_val in val_loader:
					x_val = x_val.view([batch_size, -1, n_features]).to(self.device)
					y_val = y_val.to(self.device)
					self.model.eval()
					yhat = self.model(x_val)
					val_loss = self.loss_fn(y_val, yhat).item()
					batch_val_losses.append(val_loss)
				validation_loss = np.mean(batch_val_losses)
				self.val_losses.append(validation_loss)

			if (epoch <= n_epochs) | (epoch % 50 == 0):
				print(
					f"[{epoch}/{n_epochs}] Training loss: {training_loss:.4f}\t Validation loss: {validation_loss:.4f}"
				)

		#torch.save(self.model.state_dict(), model_path)


	def evaluate(self, test_loader, batch_size=1, n_features=1, ):
		"""The method evaluate performs the model evaluation

		The method takes DataLoaders for the test dataset, batch size for mini-batch testing,
		and number of features as inputs. Similar to the model validation, it iteratively
		predicts the target values and calculates losses. Then, it returns two lists that
		hold the predictions and the actual values.

		Note:
			This method assumes that the prediction from the previous step is available at
			the time of the prediction, and only does one-step prediction into the future.

		Args:
			test_loader (torch.utils.data.DataLoader): DataLoader that stores test data
			batch_size (int): Batch size for mini-batch training
			n_features (int): Number of feature columns

		Returns:
			list[float]: The values predicted by the model
			list[float]: The actual values in the test set.

		"""
		with torch.no_grad():
			predictions = []
			values = []
			for x_test, y_test in test_loader:
				x_test = x_test.view([batch_size, -1, n_features]).to(self.device)
				y_test = y_test.to(self.device)
				self.model.eval()
				yhat = self.model(x_test)
				#predictions.append(yhat.to(self.device).detach().numpy())
				predictions.append(yhat.cpu().detach().numpy())
				#values.append(y_test.to(self.device).detach().numpy())
				values.append(y_test.cpu().detach().numpy())

		return predictions, values

	def predict(self, forecast_loader, batch_size=1, n_features=1, ):
		"""The method performs the model forecasting

		The method takes DataLoaders for the forecast dataset, batch size for mini-batch testing,
		and number of features as inputs. Similar to the model validation, it iteratively
		predicts the target values and calculates losses. Then, it returns two lists that
		hold the predictions and the actual values.

		Note:
			This method assumes that the prediction from the previous step is available at
			the time of the prediction, and only does one-step prediction into the future.

		Args:
			forecast_loader (torch.utils.data.DataLoader): DataLoader that stores future forecast data
			batch_size (int): Batch size for mini-batch training
			n_features (int): Number of feature columns

		Returns:
			list[float]: The values predicted by the model

		"""
		with torch.no_grad():
			predictions = []
			for x in forecast_loader:
				x = x.view([batch_size, -1, n_features]).to(self.device)
				self.model.eval()
				yhat = self.model(x)
				#predictions.append(yhat.to(self.device).detach().numpy())
				predictions.append(yhat.cpu().detach().numpy())
				
		return predictions

	# def predict(self, test_loader, batch_size,  n_features=1,):
	# 	"""
	# 	Window-step prediction into the future 
	# 	"""
	# 	#self.model.eval()
	# 	#for i in range(future_prediction):

	# 	with torch.no_grad():

	# 		predictions_forecast_window = list()
	# 		for i in range(future_prediction):


	# 	# 	for x_test, y_test in test_loader:
	# 	# 		x_test = x_test.view([batch_size, -1, n_features]).to(self.device)
	# 	# 		self.model.eval()
	# 	# 		yhat = self.model(x_test)
	# 	# 		predictions_forecast_window.append(yhat.to(self.device).detach().numpy())
	# 	# return predictions_forecast_window


	def forecast_torch(self, full_loader,  n_steps:int, batch_size=1, n_features=1,):

		predictions = []

		for x, y in full_loader:

			x = torch.roll(x, shifts=1, dims=2)
			x[..., -1, 0] = torch.Tensor(np.array(y)).item(0)
			with torch.no_grad():

				self.model.eval()
				for _ in range(n_steps):
					x = x.view([batch_size, -1, n_features]).to(device)
					yhat = self.model(x)
					yhat = yhat.to(device).detach().numpy()
					x = torch.roll(x, shifts=1, dims=2)
					x[..., -1, 0] = yhat.item(0)
					predictions.append(yhat)

		return predictions


	def forecast(self, full_loader,  n_steps:int, batch_size=1, n_features=1,):
		predictions = []
		values = []
		for x, y in full_loader:
			for _ in range(n_steps):
				x = x.view([batch_size, -1, n_features]).to(self.device)
				y = y.to(self.device)
				self.model.eval()
				yhat = self.model(x)
				#yhat = yhat.to(self.device).detach().numpy()
				x = torch.roll(x, shifts=1, dims=2)
				x[..., -1, 0] = yhat.item() #0
				predictions.append(yhat.to(self.device).detach().numpy())
				values.append(y.to(self.device).detach().numpy())

		return predictions, values


	def plot_losses(self):
		"""The method plots the calculated loss values for training and validation
		"""
		plt.plot(self.train_losses, label="Training loss")
		plt.plot(self.val_losses, label="Validation loss")
		plt.legend()
		plt.title("Losses")
		plt.show()
		plt.close()




#! /usr/bin/env python
"""
@author: Ajay Arunachalam
Created on: 04/10/2021
Goal: Data Preprocessing for Time-series data 
Version: 0.0.5
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
# set default figure dimensions
sns.set(rc={'figure.figsize':(11, 4)})
import pathlib
from scipy import stats
from ecgdetectors import Detectors
import warnings
warnings.filterwarnings('ignore')
import plotly.graph_objs as go
from plotly.offline import iplot

class Preprocessing:
	
	# function to find missing values
	def missing(x):
		return sum(x.isnull())

	# function to print missing in whole dataset row & column wise

	def print_missing(df):
		print("Missing information for whole dataset")
		print("Missing by row", df.apply(Preprocessing.missing, axis=0)) # col
		print("Missing by col", df.apply(Preprocessing.missing, axis=1)) # row


	# function to convert to datetime object, and extract date & time columns seperately
	def datetime_extraction(df, datetimecolindex):
		import pandas as pd
		import datetime
		# convert to datetime object

		df.iloc[:,datetimecolindex] = pd.to_datetime(df.iloc[:,datetimecolindex])

		# Extract date & time separately
		df['Date'] = df.iloc[:,datetimecolindex].dt.date
		df['Time'] = df.iloc[:,datetimecolindex].dt.time

		#Reorder the DATE, TIME column back to BEGINNING index
		cols = list(df)
		# move the column to head of list 
		cols.pop(cols.index('Date')), cols.pop(cols.index('Time'))
		df = df[['Date','Time']+cols[1:]]
		return df

	# function to join date & time columns together as single timestamp attribute

	def datetime_merge(df, datecol:str, timecol:str):
		import pandas as pd
		import datetime

		# concat date & time to form timestamp column

		if df[datecol].dtypes=='O' or df[timecol].dtypes=='O': 

			df['TS'] = df[datecol] + " " + df[timecol]
			df['TS'] = pd.to_datetime(df[datecol] + " " + df[timecol])
			# move newly create TS column back to start postion
			cols = list(df)
			cols.pop(cols.index('TS'))
			df = df[['TS'] + cols[2:]]

		else:
			df['TS'] = df.apply(lambda x : pd.datetime.combine(x['datecol'],x['timecol']),1)
			# move newly create TS column back to start postion
			cols = list(df)
			cols.pop(cols.index('TS'))
			df = df[['TS'] + cols[2:]]
		
		return df

	# impute missing values
	def impute(df, modes:int): #start_sensor_column_index, 
		import pandas as pd
		if modes==0:
			#df.iloc[:,start_sensor_column_index:len(df.columns)].fillna(0, inplace=True)
			df.fillna(0, inplace=True)
		elif modes==1:
			#df.iloc[:,start_sensor_column_index:len(df.columns)].fillna(df.mean(), inplace=True)
			df.fillna(df.mean(), inplace=True)
		elif modes==2:
			#df.iloc[:,start_sensor_column_index:len(df.columns)].fillna(df.median(), inplace=True)
			df.fillna(df.median(), inplace=True)
		else:
			#df.iloc[:,start_sensor_column_index:len(df.columns)].fillna(method='bfill', inplace=True) # backfill / bfill: use next valid observation to fill gap.
			df.fillna(method='bfill', inplace=True)
		df.to_csv('./df_no_NA.csv')
		return df

	# drop particular column
	def drop_columns(df, column_index_to_drop:int):
		df.drop(df.columns[column_index_to_drop], inplace=True, axis=1)
		return df



class ExploratoryDataAnalysis:
	
	# profiling your dataset
	def profiling(df): 
		import pandas_profiling as pp       
		# To Generate a HTML report file
		html_report = pp.ProfileReport(df, title="Data Profiling Report")
		html_report.to_file("./profiling_report.html")
	
	# filtering outliers
	def drop_outliers(df, z_threshold):
		from scipy import stats
		import numpy as np
		constraints = df.select_dtypes(include=[np.number]) \
			.apply(lambda x: np.abs(stats.zscore(x)) < z_threshold) \
			.all(axis=1)
		df.drop(df.index[~constraints], inplace=True)
		return df

	# univariate/multivariate plots
	def plots(df, cols):
		# plotting multivariate
		df[cols].plot()

	# Usage: ExploratoryDataAnalysis.plots(df=processed_data, cols=['col1', 'col2', 'col3', 'col4', 'col5'])


	# plot with plotly
	def plot_dataset(df, fc, title):
		import plotly.graph_objs as go
		from plotly.offline import iplot
		data = []
		value = go.Scatter(
			x=df.index,
			y=df.value,
			mode="lines",
			name="values",
			marker=dict(),
			text=df.index,
			line=dict(color="rgba(0,0,0, 0.3)"),
		)
		data.append(value)

		layout = dict(
			title=title,
			xaxis=dict(title=f"{df.index.name}", ticklen=5, zeroline=False),
			yaxis=dict(title=f"{fc}", ticklen=5, zeroline=False),
		)

		fig = dict(data=data, layout=layout)
		iplot(fig) 

	# Usage: 
	'''
	df = df.set_index(['your_datatime_column'])
	df = df.rename(columns={'your_column_name': 'value'})

	df.index = pd.to_datetime(df.index)
	if not df.index.is_monotonic:
		df = df.sort_index()
	
	ExploratoryDataAnalysis.plot_dataset(df, title='Enter your title of the plot')
	'''
		
	def sumplots_rw(df, cols, window:int):
		# sumplots of rolling window
		processed_data[cols].rolling(window=window).mean().plot(subplots=True)

	# Usage: ExploratoryDataAnalysis.sumplots_rw(processed_data, ['col1', 'col2', 'col3', 'col4', 'col5'], 12)

	# detects r-signal-to-signal peak
	# includes the following detector types, 
	# i.e., det_type = [swt_detector, christov_detector, two_average_detector, engzee_detector, hamilton_detector, 
	#                  pan_tompkins_detector, matched_filter_detector]
	def r_peak_plots_arg(dataframe, det_type, col_index, fs:int):
		unfiltered_ecg_dat = dataframe 
		unfiltered_ecg = unfiltered_ecg_dat.iloc[:,col_index]
		fs = fs
		detectors = Detectors(fs)
		det_type = str(det_type.strip('\"'))
		#print(det_type)
		detector_call = getattr(detectors, det_type);
		r_peaks = detector_call(unfiltered_ecg)
		plt.figure()
		plt.plot(unfiltered_ecg)
		plt.plot(r_peaks, unfiltered_ecg[r_peaks], 'ro')
		plt.title('Detected R-peaks')
		plt.show()
		return det_type, col_index 

	# Usage: ExploratoryDataAnalysis.r_peak_plots_arg(processed_data,det_type='christov_detector', col_index=6, fs=1000)
	

class DescriptiveStatistics:

	# Stats for whole dataset
	def desc_all(df, start_sensor_column_index):
		print("Descriptive Statistics")
		return df.iloc[:,start_sensor_column_index:len(df.columns)].describe() # include='all'
	
	# Stats with relation to response variable and any feature (tobe used if your data is labeled)
	def desc_target_feature(df, target, feature):
		print("Descriptive Statistics by Target-Feature")
		return df.groupby(target)[feature].describe()


class StatisticalEvaluation:

	# Test your data for normality/if the data follows normal distribution
	def normality_test(df, start_sensor_column_index):
		from scipy import stats, norm, shapiro, jarque_bera
		print("Shapiro-Wilk Normality test (W test statistic, the p-value) =======> {}  ".format(shapiro(df.iloc[:,start_sensor_column_index:len(df.columns)])))
		print("Jarque-Bera Normality test (W test statistic, the p-value) =======> {}".format(jarque_bera(df.iloc[:,start_sensor_column_index:len(df.columns)])))


class Features:

	def generate_time_lags(df, n_lags:int, col_name:str):
		df_n = df.copy()
		for n in range(1, n_lags + 1):
			df_n[f"lag{n}"] = df_n[col_name].shift(n) 
		df_n = df_n.iloc[n_lags:]
		return df_n

	def generate_date_time_features_hour(df, cols:list()):

		df_features = (
				df
				.assign(hour = df.index.hour)
				.assign(day = df.index.day)
				.assign(month = df.index.month)
				.assign(day_of_week = df.index.dayofweek)
				.assign(week_of_year = df.index.week)
		)
				
		return df_features

	def generate_date_time_features_month(df, cols:list()):

		df_features = (
				df
				.assign(month = df.index.month)
				.assign(day_of_week = df.index.dayofweek)
				.assign(week_of_year = df.index.week)
		)
				
		return df_features

	def generate_date_time_features_ohe(df, cols:list()):

		df_features = (
				df
				.assign(hour = df.index.hour)
				.assign(day = df.index.day)
				.assign(month = df.index.month)
				.assign(day_of_week = df.index.dayofweek)
				.assign(week_of_year = df.index.week)
		)
		df_features = df_features

		def onehot_encode_pd(df_features=df_features, cols=cols):
			import pandas as pd
			df_with_dummies = pd.get_dummies(df_features, columns=cols)
			df_with_features_dummies = pd.concat([df_features, df_with_dummies.iloc[:,1:]], axis=1)
			return df_with_features_dummies

		df_full_features = onehot_encode_pd(df_features=df_features, cols=cols)
		
		return df_full_features


	def generate_cyclic_features(df, col_name, period, start_num=0):
		kwargs = {
			f'sin_{col_name}' : lambda x: np.sin(2*np.pi*(df[col_name]-start_num)/period),
			f'cos_{col_name}' : lambda x: np.cos(2*np.pi*(df[col_name]-start_num)/period)    
			 }
		df_full_features = df.assign(**kwargs).drop(columns=[col_name])
		return df_full_features



	def generate_other_related_features(df):
		from datetime import date
		import holidays

		country_holidays = holidays.US() # change according to user specific country

		def is_holiday(date):
			date = date.replace(hour = 0)
			return 1 if (date in country_holidays) else 0

		def add_holiday_col(df, holidays):
			return df.assign(is_holiday = df.index.to_series().apply(is_holiday))

		df_full_features = add_holiday_col(df=df, holidays=country_holidays)
		
		return df_full_features







#! /usr/bin/env python

"""
@author: Ajay Arunachalam
Created on: 25/10/2021
Training the forecasting and Nowcasting model
Version: 0.0.5
"""


import torch
from torch.utils.data import TensorDataset, DataLoader
import torch.optim as optim
from .utility import *
from .dpp import *
#from .forecast_ml import *
from .forecast_ml_extension import *
from .denoise import *
from .similarity import *
from .gnn_layer import *
from .stats import *

class Forecast:

	ts = globals()
	fc = globals()

	select_model = globals() # Possible values ['rnn','lstm', 'gru', 'em', etc]
	select_user_path = globals() # Provide user_path './forecast_folder/'
	select_scaler = globals() # Possible values ['minmax','standard','maxabs','robust']
	forecast_window = globals() # no. of timesteps/points to be used for the forecasting model and nowcasting period

	hidden_dim = globals()
	layer_dim = globals()
	batch_size = globals()
	dropout = globals()
	n_epochs = globals()
	learning_rate = globals()
	weight_decay = globals()


	def set_variable(**kwargs):
		for key, value in kwargs.items():
			print("{0} = {1}" .format(key,value))

		ts = list(kwargs.values())[0]
		fc = list(kwargs.values())[1]

		return ts, fc

	assert ts == ts
	assert fc == fc

	def set_model_config(**kwargs):

		for key, value in kwargs.items():
			print("{0} = {1}" .format(key,value))

		select_model = list(kwargs.values())[0]
		select_user_path = list(kwargs.values())[1]
		select_scaler = list(kwargs.values())[2]
		forecast_window = list(kwargs.values())[3]

		return select_model, select_user_path, select_scaler, forecast_window

	assert select_model == select_model
	assert select_user_path == select_user_path
	assert select_scaler == select_scaler
	assert forecast_window ==  forecast_window

	def hyperparameter_config(**kwargs):

		for key, value in kwargs.items():
			print("{0} = {1}" .format(key,value))

		hidden_dim = list(kwargs.values())[0]
		layer_dim = list(kwargs.values())[1]
		batch_size = list(kwargs.values())[2]
		dropout = list(kwargs.values())[3]
		n_epochs = list(kwargs.values())[4]
		learning_rate = list(kwargs.values())[5] 
		weight_decay = list(kwargs.values())[6]

		return hidden_dim, layer_dim, batch_size, dropout, n_epochs, learning_rate, weight_decay

	assert hidden_dim == hidden_dim
	assert layer_dim == layer_dim
	assert batch_size == batch_size
	assert dropout == dropout
	assert n_epochs == n_epochs
	assert learning_rate == learning_rate
	assert weight_decay == weight_decay

	def forecast(df, ts, fc, opt, scaler, period:int, fq:str, select_scaler=select_scaler, ):

		digit = ''.join(filter(str.isdigit, str(fq)))
		interval = int(digit)

		ff_df = Helper.make_future_df(df, ts, period, fq)

		print(f'Forecast period dataframe: {ff_df.index}')

		#print(f'Forecast period dataframe: {ff_df.index.hour}')

		#cols=['hour','month','day','day_of_week','week_of_year']

		frequency = ''.join([i for i in fq if not i.isdigit()])

		#if any(str(frequency).startswith(tuple(l)) for l in ['h', 'H', 's', 'S', 'min', 'MIN', 'n', 'N']) and interval is not None:
		if frequency in ['h', 'H', 's', 'S', 'min', 'MIN', 'n', 'N'] and interval is not None:


		#if str(fq)=='h' or str(fq)=='H' or str(fq) == 's' or str(fq) == 'S' or str(fq) == 'min' or str(fq) == 'MIN' or str(fq) == 'n' or str(fq) == 'N' or interval is not None:  # hourly(h:m:s)

			ff_full_features = Features.generate_date_time_features_hour(ff_df, ['hour','month','day','day_of_week','week_of_year'])
			ff_full_features = Features.generate_cyclic_features(ff_full_features, 'hour', 24, 0)
			ff_full_features = Features.generate_cyclic_features(ff_full_features, 'day_of_week', 7, 0)
			ff_full_features = Features.generate_cyclic_features(ff_full_features, 'month', 12, 1)
			ff_full_features = Features.generate_cyclic_features(ff_full_features, 'week_of_year', 52, 0)

			ff_full_features = Features.generate_other_related_features(df=ff_full_features)

		elif frequency in ['d', 'D', 'w', 'W', 'm', 'M', 'q', 'Q', 'QS', '2q', '2Q', 'HA', 'Y', 'y', 'A'] and interval is not None:

		#elif any(str(frequency).startswith(tuple(l)) for l in ['d', 'D', 'w', 'W', 'm', 'M', 'q', 'Q', 'QS', '2q', '2Q', 'HA', 'Y', 'y', 'A']) and interval is not None:

		#elif str(fq) == 'd' or str(fq) == 'D' or str(fq) == 'w' or str(fq) == 'W' or str(fq)=='m' or str(fq)=='M' or str(fq) == 'q' or str(fq) == 'Q' or str(fq) == 'QS' or str(fq) == '2q' or str(fq) == '2Q' or str(fq) == 'HA' or str(fq) == 'y' or str(fq) == 'Y' or str(fq) == 'A' or interval is not None: # Yearly(Daily, weekly, monthly, quater, semi-annual, annual)

			ff_full_features = Features.generate_date_time_features_month(ff_df, ['month','day_of_week','week_of_year'])
			ff_full_features = Features.generate_cyclic_features(ff_full_features, 'day_of_week', 7, 0)
			ff_full_features = Features.generate_cyclic_features(ff_full_features, 'month', 12, 1)
			ff_full_features = Features.generate_cyclic_features(ff_full_features, 'week_of_year', 52, 0)

			ff_full_features = Features.generate_other_related_features(df=ff_full_features)


		X = ff_full_features
		
		input_dim = len(X.columns)
		#X, y = Helper.predictor_outcome_split(df, target_col)
		X_arr = Helper.apply_transformation_forecast(X, select_scaler)

		unseen_loader = Helper.prepare_pytorch_data_forecast_df(X_arr)

		predictions = opt.predict(
			unseen_loader,
			batch_size=1,
			n_features=input_dim
		)

		ff_result = Helper.forecast_window_inference(predictions, ff_df, scaler)
		print(f'Forecast period predictions: {ff_result}')

		Helper.plot_forecast(ff_result, fc)

		forecasted_data = Helper.save_final_data(df, ff_result, ts, fc)
		ff_full_features_ = pd.concat([ff_result, ff_full_features], axis=1)
		return forecasted_data, ff_full_features, ff_full_features_


	def train(df, target_col, split_ratio:float, select_model=select_model, select_scaler=select_scaler, forecast_window=forecast_window, hidden_dim=hidden_dim, layer_dim=layer_dim, batch_size=batch_size,dropout=dropout, n_epochs=n_epochs, learning_rate=learning_rate, weight_decay=weight_decay):
		from torch.utils.data import TensorDataset, DataLoader

		X_train, X_val, X_test, y_train, y_val, y_test = Helper.train_val_test_split(df, target_col, split_ratio) #'value', 0.2
		X_train_arr, X_val_arr, X_test_arr, y_train_arr, y_val_arr, y_test_arr, scaler = Helper.apply_transformation(X_train, X_val, X_test, y_train, y_val, y_test, select_scaler)
		
		train_loader, val_loader, test_loader, test_loader_one = Helper.prepare_pytorch_data(X_train_arr, X_val_arr, X_test_arr, y_train_arr, y_val_arr, y_test_arr, batch_size=batch_size)
		'''

		scaler = Helper.get_scaler(str(select_scaler)) #'minmax'
		X_train_arr = scaler.fit_transform(X_train)
		X_val_arr = scaler.transform(X_val)
		X_test_arr = scaler.transform(X_test)

		y_train_arr = scaler.fit_transform(y_train)
		y_val_arr = scaler.transform(y_val)
		y_test_arr = scaler.transform(y_test)

		#batch_size = 64

		train_features = torch.Tensor(X_train_arr)
		train_targets = torch.Tensor(y_train_arr)
		val_features = torch.Tensor(X_val_arr)
		val_targets = torch.Tensor(y_val_arr)
		test_features = torch.Tensor(X_test_arr)
		test_targets = torch.Tensor(y_test_arr)

		train = TensorDataset(train_features, train_targets)
		val = TensorDataset(val_features, val_targets)
		test = TensorDataset(test_features, test_targets)

		train_loader = DataLoader(train, batch_size=batch_size, shuffle=False, drop_last=True)
		val_loader = DataLoader(val, batch_size=batch_size, shuffle=False, drop_last=True)
		test_loader = DataLoader(test, batch_size=batch_size, shuffle=False, drop_last=True)
		test_loader_one = DataLoader(test, batch_size=1, shuffle=False, drop_last=True)
		'''

		input_dim = len(X_train.columns)
		# output_dim = 1
		# hidden_dim = 64
		# layer_dim = 3
		# batch_size = 64
		# dropout = 0.2
		# n_epochs = 5
		# learning_rate = 1e-3
		# weight_decay = 1e-6

		output_dim = 1
		hidden_dim = hidden_dim
		layer_dim = layer_dim
		batch_size = batch_size
		dropout = dropout
		n_epochs = n_epochs
		learning_rate = learning_rate
		weight_decay = weight_decay


		model_params = {'input_dim': input_dim,
						'hidden_dim' : hidden_dim,
						'layer_dim' : layer_dim,
						'output_dim' : output_dim,
						'dropout_prob' : dropout}

		model = Helper.get_model(str(select_model), model_params) # 'lstm'

		loss_fn = nn.MSELoss(reduction="mean")
		optimizer = optim.Adam(model.parameters(), lr=learning_rate, weight_decay=weight_decay)
		device = "cuda" if torch.cuda.is_available() else "cpu"
		print(f"{device}" " is available.")


		opt = Optimization(device=device, model=model, loss_fn=loss_fn, optimizer=optimizer)
		opt.train(train_loader, val_loader, batch_size=batch_size, n_epochs=n_epochs, n_features=input_dim)
		opt.plot_losses()

		predictions, values = opt.evaluate(
			test_loader_one,
			batch_size=1,
			n_features=input_dim
		)

		#scaler = Helper.get_scaler(select_scaler)

		df_result = Helper.format_predictions(predictions, values, X_test, scaler)
		print(f'Forecast testset predictions: {df_result}')

		result_metrics,  key_metrics = Helper.calculate_metrics(df_result)
		print(f'Model Evaluations: {result_metrics}')

		print(f'Model Evaluations: {key_metrics}')

		Helper.plot_metrics(result_metrics, key_metrics)

		df_baseline = Helper.build_baseline_model(df, split_ratio, target_col) #df_feature, 0.2, 'value'
		baseline_metrics = Helper.calculate_metrics(df_baseline)

		Helper.plot_predictions(df_result, df_baseline)

		return opt, scaler


	









# To get modules
from . import main
from . import dpp 
from . import forecast_ml 
from . import forecast_ml_extension
from . import stats
from . import utility 
from . import denoise
from . import similarity

# To get sub-modules
from .main import *
from .dpp import *
from .forecast_ml import *
from .forecast_ml_extension import *
from .stats import *
from .utility import *
from .denoise import *
from .similarity import *

#################################################################################
module_type = 'Running' if  __name__ == "__main__" else 'Imported'
version_number = '0.0.5'
print(f"""{module_type} DeepXF version:{version_number}. Example call by using:

******************   SET FORECAST MODEL/BASE CONFIGURATIONS   ********************

# select algorithms, scaler, etc.
select_model, select_user_path, select_scaler, forecast_window = Forecast.set_model_config(select_model='rnn', select_user_path='./forecast_folder_path/', select_scaler='minmax', forecast_window=1)

----------------------------------------------------------------------------------

# set variables
ts, fc = Forecast.set_variable(ts='Datetime', fc='PJME_MW')

# get variables
model_df, orig_df = Helper.get_variable(df, ts, fc)

----------------------------------------------------------------------------------

# set hyperparameters
hidden_dim, layer_dim, batch_size, dropout, n_epochs, learning_rate, weight_decay = Forecast.hyperparameter_config(hidden_dim=64,layer_dim = 3, batch_size=64, dropout = 0.2, n_epochs = 30, learning_rate = 1e-3, weight_decay = 1e-6)

----------------------------------------------------------------------------------

*****************   Calling DEEP LEARNING BASED FORECASTING   *********************

# train selected deep learning model
opt, scaler = Forecast.train(df=df_full_features, target_col='value', split_ratio=0.2, select_model=select_model, select_scaler=select_scaler, forecast_window=forecast_window, batch_size=batch_size, hidden_dim=hidden_dim, layer_dim=layer_dim,dropout=dropout, n_epochs=n_epochs, learning_rate=learning_rate, weight_decay=weight_decay)

----------------------------------------------------------------------------------

# forecast for the user provided window
forecasted_data, ff_full_features, ff_full_features_ = Forecast.forecast(model_df, ts, fc, opt, scaler, period=25, fq='1h', select_scaler=select_scaler,)

---------------------------------------------------------------------------------

# model interpretation
Helper.explainable_forecast(df_full_features, ff_full_features_, fc, specific_prediction_sample_to_explain=df.shape[0]+2, input_label_index_value=0, num_labels=1)


******************   DYNAMIC FACTOR MODEL BASED NOWCASTING   *********************

# select algorithms, scaler, etc.
select_model, select_user_path, select_scaler, forecast_window = Forecast.set_model_config(select_model='em', select_user_path='./forecast_folder_path/', select_scaler='minmax', forecast_window=5)

----------------------------------------------------------------------------------

# train nowcasting model with EM-algorithm
nowcast_full_data, nowcast_pred_data = EMModel.nowcast(df_full_features, ts, fc, period=5, fq='1h', forecast_window=forecast_window, select_model=select_model)

----------------------------------------------------------------------------------

# get interpretable results
EMModel.explainable_nowcast(df_full_features, nowcast_pred_data, fc, specific_prediction_sample_to_explain=df.shape[0]+2, input_label_index_value=0, num_labels=1)

***********************   DENOISING TIMESERIES ECG SIGNALS   **********************

# set thresholds
fs, order, cutoff_high, cutoff_low, powerline, nyq = Denoise.set_parameters(fs = 500, order = 5, cutoff_high = 0.5, cutoff_low = 2, powerline = 60)

# filter noise and save output
Denoise.plot_and_write(ecg_signal_data, featurecol_index=2, fs=fs, cutoff_high=cutoff_high, cutoff_low=cutoff_low, nyq = nyq, powerline=powerline, order=order)


***********************   TIMESERIES ECG SIGNAL SIMILARITY   **********************

# configure, and train siamese neural network
sm = Siamese.Siamese_Model(5) # 5 - # number of features
sm.compile(loss='mse', optimizer='adam', metrics=['mse', 'mae', 'mape', 'cosine'])
score = sm(inputs=[np.array(activity_input_sequence_1), np.array(activity_input_sequence_2)]) # signal similarity scores

""")







#! /usr/bin/env python

"""
@author: Ajay Arunachalam
Created on: 12/10/2021
Goal: Graph Neural Network Layers
Version: 0.0.5
"""
import numpy as np
import torch
from torch import nn
import torch.nn.functional as F

def device_as(x,y):
	return x.to(y.device)

# tensor operations now support batched inputs
def calc_degree_matrix_norm(a):
	return torch.diag_embed(torch.pow(a.sum(dim=-1),-0.5))

def create_graph_lapl_norm(a):
	size = a.shape[-1]
	a +=  device_as(torch.eye(size),a)
	D_norm = calc_degree_matrix_norm(a)
	L_norm = torch.bmm( torch.bmm(D_norm, a) , D_norm )
	return L_norm

class GCN_Layer(nn.Module):
		"""
		A simple GCN layer, similar to https://arxiv.org/abs/1609.02907
		"""
		def __init__(self, in_features, out_features, bias=True):

			super().__init__()
			self.linear = nn.Linear(in_features, out_features, bias=bias)

		def forward(self, X, A):
				"""
				A: adjecency matrix
				X: graph signal
				"""
				L = create_graph_lapl_norm(A)
				x = self.linear(X)
				return torch.bmm(L, x)


#! /usr/bin/env python
"""
@author: Ajay Arunachalam
Created on: 11/10/2021
Goal: Denoising time-series signals
Version: 0.0.5
"""
from scipy import signal
from scipy.signal import butter, iirnotch, lfilter
import numpy as np 
import pandas as pd
import matplotlib.pyplot as plt

class Denoise:

	fs = globals()
	order = globals()
	cutoff_high = globals()
	cutoff_low = globals()
	powerline = globals()
	nyq = globals()

	def set_parameters(**kwargs):
		for key, value in kwargs.items():
			print("{0} = {1}" .format(key, value))

		fs = list(kwargs.values())[0]
		order = list(kwargs.values())[1]
		cutoff_high = list(kwargs.values())[2]
		cutoff_low = list(kwargs.values())[3]
		powerline = list(kwargs.values())[4]

		nyq = 0.5 * fs

		return fs, order, cutoff_high, cutoff_low, powerline, nyq

	assert fs == fs
	assert order == order
	assert cutoff_high == cutoff_high
	assert cutoff_low == cutoff_low
	assert powerline == powerline

	assert nyq == nyq

	## Butter high pass filter allows frequencies higher than a cut-off value
	def bhpf(cutoff, order, nyq):
		#nyq = 0.5 * fs
		normal_cutoff = cutoff/nyq
		b, a = butter(order, normal_cutoff, btype='high', analog=False, output='ba')
		return b, a

	## Butter low pass filter allows frequencies lower than a cut-off value
	def blpf(cutoff, order, nyq):
		#nyq = 0.5 * fs
		normal_cutoff = cutoff/nyq
		b, a = butter(order, normal_cutoff, btype='low', analog=False, output='ba')
		return b, a

	## Notch filter 
	def nf(cutoff, q, nyq):
		#nyq = 0.5 * fs
		freq = cutoff/nyq
		b, a = iirnotch(freq, q)
		return b, a

	## Applying high pass filter 
	def hpfd(data, cutoff_high, order, nyq):
		b,a = bhpf(cutoff_high, order, nyq)
		x = lfilter(b,a,data)
		return x

	## Applying low pass filter 
	def lpfd(data, cutoff_low, order, nyq):
		b,a = blpf(cutoff_low, order, nyq)
		y = lfilter(b,a,data)
		return y

	## Applying notch filter
	def nd(data, powerline, q):
		b,a = nf(powerline,q)
		z = lfilter(b,a,data)
		return z

	## Ensembling of the filter
	def overlay_filter(data, cutoff_high, cutoff_low, nyq, powerline, fs, order):
		b, a = Denoise.bhpf(cutoff_high, order, nyq)
		x = lfilter(b, a, data)
		d, c = Denoise.blpf(cutoff_low, order, nyq)
		y = lfilter(d, c, x)
		f, e = Denoise.nf(powerline, 30, nyq)
		z = lfilter(f, e, y)     
		return z

	def plot_and_write(df, featurecol_index:int, fs=fs, cutoff_high=cutoff_high, cutoff_low=cutoff_low, nyq=nyq, powerline=powerline, order=order):
		#fs = 500; cutoff_high = 0.5; cutoff_low = 2; powerline = 60
		#order = 5 ## Order of five works well with ECG signals

		plt.figure(1)
		ax1 = plt.subplot(121)
		plt.plot(df.iloc[:,featurecol_index:])
		ax1.set_title("Raw TimeSeries ECG Signal")

		filter_ts_signal = Denoise.overlay_filter(df.iloc[:,featurecol_index:], cutoff_high, cutoff_low, nyq, powerline, fs, order)
		ax2 = plt.subplot(122)
		plt.plot(filter_ts_signal)
		ax2.set_title("Denoised TimeSeries ECG Signal")
		plt.show()
		#filtered_dataset = pd.concat([df.iloc[:,:featurecol_index], filter_ts_signal])
		cols = list(df.iloc[:,featurecol_index:])
		#filtered_dataset = pd.DataFrame({'Date':df.Date, 'Time': df.Time,'ecg1': filter_ts_signal[:, 0], 'ecg2': filter_ts_signal[:, 1],'ecg3': filter_ts_signal[:, 2],'ecg4': filter_ts_signal[:, 3],'ecg5': filter_ts_signal[:, 4]})
		filter_ts_signal_dataset = pd.DataFrame(np.array(filter_ts_signal), columns=cols) 
		filtered_dataset = pd.concat([df.iloc[:, :featurecol_index].reset_index(), filter_ts_signal_dataset.reset_index()], axis=1)
		filtered_dataset.drop(columns=['index'], inplace=True)
		filtered_dataset.to_csv('Denoise_TS_Signal.csv', index=None)
		return filtered_dataset

	def write_csv():
		pass
		





#! /usr/bin/env python

"""
@author: Ajay Arunachalam
Created on: 25/10/2021
Goal: Explainable Nowcasting with Dynamic Factor Model based on EM algorithm
Version: 0.0.5
"""

import statsmodels.api as sm
import pandas as pd
import plotly.offline as pyo
import plotly.graph_objs as go
from plotly.offline import iplot
import matplotlib.pyplot as plt

class EMModel():
	'''
	Expectation–maximization algorithm
	'''
	def __init__(self, df, model, FORECAST_PERIOD):
		super(EMModel, self).__init__()
		self.df = df
		self.fp = FORECAST_PERIOD
		self.mod = model
	def predict(self):
		self.mod = sm.tsa.DynamicFactorMQ(self.df)
		print(self.mod.summary())
		self.res = self.mod.fit(disp=5)
		#print(self.res.forecast(steps=self.fp))
		pred = self.res.forecast(steps=self.fp)
		return pd.DataFrame(pred)

	def get_stats_model(model):
		models = {
			"em": EMModel,

		}
		return models.get(model.lower())


	# old

	# def make_future_df(df, ts, period:int, fq:str):
	# 	if str(fq) == 'h' or str(fq) == 'H':
	# 		#last_date = df[ts].max()
	# 		last_date = df.index.max()
	# 		dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
	# 		dates = dates[dates > last_date]  # Drop start if equals last_date
	# 		dates = dates[:period]  # Return correct number of periods
	# 		ff = pd.DataFrame({ts:dates})
	# 		ff.set_index(ts, inplace=True)
	# 	elif str(fq) == 'm' or str(fq) == 'M':
	# 		last_date = df.index.max()
	# 		dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
	# 		dates = dates[dates > last_date]  # Drop start if equals last_date
	# 		dates = dates[1:period+1]  # Return correct number of periods
	# 		ff = pd.DataFrame({ts:dates})
	# 		ff.set_index(ts, inplace=True)
	# 	return ff

	# revised

	def make_future_df(df, ts, period:int, fq:str):
		#last_date = df.index.max()
		#dates = pd.date_range(start=last_date, periods=period + 1, freq=fq)
		#assert dates == dates
		digit = ''.join(filter(str.isdigit, str(fq)))
		interval = int(digit)
		if str(fq) == 'h' or str(fq) == 'H' or interval is not None:  # hourly
			#last_date = df[ts].max()
			last_date = df.index.max()
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[:period]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		elif str(fq) == 'm' or str(fq) == 'M' or interval is not None: # monthly
			last_date = df.index.max()
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[1:period+1]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		elif str(fq) == 'd' or str(fq) == 'D' or interval is not None: # daily
			last_date = df.index.max()
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[1:period+1]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		elif str(fq) == 'w' or str(fq) == 'W' or interval is not None: # weekly
			last_date = df.index.max()
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[1:period+1]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		elif str(fq) == 'min' or str(fq) == 'MIN' or interval is not None: # minute
			last_date = df.index.max()
			diff_min = (df.index.minute[1] - df.index.minute[0])
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[:period]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		elif str(fq) == 's' or str(fq) == 'S' or interval is not None: # second
			last_date = df.index.max()
			diff_sec = (df.index.second[1] - df.index.second[0])
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[:period]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		elif str(fq) == 'n' or str(fq) == 'N' or interval is not None: # millisecond
			last_date = df.index.max()
			diff_sec = (df.index.second[1] - df.index.second[0])
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[:period]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		elif str(fq) == 'q' or str(fq) == 'Q' or str(fq) == 'QS': # quaterly
			last_date = df.index.max()
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[:period]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		elif str(fq) == '2q' or str(fq) == '2Q' or str(fq) == 'HA': # half yearly
			last_date = df.index.max()
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[:period]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		elif str(fq) == 'y' or str(fq) == 'Y' or str(fq) == 'A' or interval is not None: # yearly
			last_date = df.index.max()
			dates = pd.date_range(start=last_date, periods=period + 1, freq=fq) # Add 1 incase last date is included
			dates = dates[dates > last_date]  # Drop start if equals last_date
			dates = dates[:period]  # Return correct number of periods
			ff = pd.DataFrame({ts:dates})
			ff.set_index(ts, inplace=True)
		return ff

	def plot_nowcast(ff, fc, FORECAST_PERIOD):
		# Create the plotly figure
		index = ff.columns.get_loc(fc)
		actual = go.Scatter(
		  x=ff.index[:-FORECAST_PERIOD],
		  y=ff.iloc[:-FORECAST_PERIOD,index],
		  mode = 'markers',
		  marker = {
			'color': '#FF0000',
			'size': 2,
			'line': {
			  'color': '#FF0000',
			  'width': 3
			}
		  },
		  name = 'Actual'
		)

		nowcast = go.Scatter(
		x=ff.index[-FORECAST_PERIOD:],
		  y=ff.iloc[-FORECAST_PERIOD:, index],
		  mode = 'markers',
		  marker = {
		'color': '#3bbed7',
		'size': 4,
		  },
		  line = {
			'width': 3
		  },
		  name = 'Nowcast',
		)

		data = [actual, nowcast]

		layout = dict(
			title="Nowcast vs Historical Values for the dataset",
				xaxis=dict(title=f"{ff.index.name}", ticklen=5, zeroline=False),
				yaxis=dict(title=f'{fc}', ticklen=5, zeroline=False),
		)

		fig = dict(data=data, layout=layout)
		iplot(fig)



	def nowcast(df, ts, fc, period:int, fq:str, forecast_window, select_model):

		ff_df = EMModel.make_future_df(df, ts, period, fq)
		model = EMModel.get_stats_model(str(select_model)) # 'em'

		em = EMModel(df=df, model=model, FORECAST_PERIOD=forecast_window)
		em_data = em.predict()
		merged_data = pd.concat([ff_df.reset_index(), em_data.reset_index()], axis=1)
		merged_data = merged_data.drop(columns=['index'])
		merged_data = merged_data.set_index(ts)
		#print(f' Merged: {merged_data}')
			
		nowcast_full_data = pd.concat([df.reset_index(), merged_data.reset_index()], axis=0)
		nowcast_full_data = nowcast_full_data.set_index(ts)
		nowcast_full_data = nowcast_full_data.rename(columns={'value':fc})
		#print(f'Nowcast Merged: {nowcast_full_data}')
		EMModel.plot_nowcast(nowcast_full_data, fc, FORECAST_PERIOD=period)
		nowcast_full_data.to_csv('./nowcast_full_data.csv', encoding='utf-8')

		return nowcast_full_data, merged_data

	def explainable_nowcast(df, ff, fc, specific_prediction_sample_to_explain:int, input_label_index_value, num_labels:int):

		"""
		Understand, interpret, and trust the results of the deep learning models at individual/samples level and multiple columns
		"""
		import shap
		import numpy as np
		import pandas as pd
		from keras.models import Sequential
		from keras.layers import Dense
		import ipywidgets as widgets

		data = pd.concat([df, ff], axis=0)
		data.rename(columns={"value":fc}, inplace=True)

		def X_y_split(df, target_col):
			y = df[[target_col]]
			X = df.drop(columns=[target_col])
			return X, y

		X, Y = X_y_split(data, fc)

		# Get the number of inputs and outputs from the dataset
		n_inputs, n_outputs = X.shape[1], Y.shape[1]

		def get_model(n_inputs, n_outputs):
			model_nn = Sequential()
			model_nn.add(Dense(32, input_dim=n_inputs, kernel_initializer='he_uniform', activation='relu'))
			model_nn.add(Dense(n_outputs, kernel_initializer='he_uniform'))
			model_nn.compile(loss='mae', optimizer='adam')
			return model_nn

		model_nn = get_model(n_inputs, n_outputs)

		# model_nn.fit(X.iloc[10:,:].values, Y, epochs=30)

		# model_nn.evaluate(x = X.iloc[10:,:].values, y = Y)

		model_nn.fit(X.iloc[:df.shape[0],:].values, Y.iloc[:df.shape[0]], epochs=30)

		model_nn.evaluate(x = X.iloc[:df.shape[0],:].values, y = Y.iloc[:df.shape[0]])

		XpredictInputData = X.iloc[specific_prediction_sample_to_explain,:] # X[specific_prediction_sample_to_explain,:]

		if (XpredictInputData.ndim == 1):
			XpredictInputData = np.array([XpredictInputData])

		print(model_nn.predict(XpredictInputData)) # 0:1

		'''
		Here we take the Keras model trained above and explain why it makes different predictions on individual samples.

		Set the explainer using the Kernel Explainer (Model agnostic explainer method form SHAP).
		'''
		explainer = shap.KernelExplainer(model = model_nn.predict, data = X.head(50), link = "identity") # data = X[0:50]

		'''
		Get the Shapley value for a single example.
		'''
		# Set the index of the specific example to explain

		shap_value_single = explainer.shap_values(X = X.iloc[specific_prediction_sample_to_explain,:], nsamples = 100)  # X[specific_prediction_sample_to_explain,:]

		'''
		Display the details of the single example
		'''
		print(X.iloc[specific_prediction_sample_to_explain,:]) 
		'''
		Choose the label/output/target to run individual explanations on:

		Note: The dropdown menu can easily be replaced by manually setting the index on the label to explain.
		'''
		# Create the list of all labels for the drop down list
		#label_cols = ['window_diff_0', 'window_diff_1', 'window_diff_2', 'window_diff_3', 'window_diff_4', 'window_diff_5', 'window_diff_6']

		#label_cols = ['window_diff_'+str(i) for i in range(num_labels)]
		label_cols = [f'{fc}_'+str(i) for i in range(num_labels)]
		#print(label_cols)
		df_labels = pd.DataFrame(data = Y, columns = label_cols)
		#df_labels.to_csv('./y_labels.csv')
		list_of_labels = df_labels.columns.to_list() # Y.columns.to_list()

		# Create a list of tuples so that the index of the label is what is returned
		tuple_of_labels = list(zip(list_of_labels, range(len(list_of_labels))))

		# Create a widget for the labels and then display the widget
		current_label = widgets.Dropdown(options=tuple_of_labels,
									  value=input_label_index_value,
									  description='Select Label:'
									  )

		# Display the dropdown list (Note: access index value with 'current_label.value')
		print(current_label)
		#Dropdown(description='Select Label:', options=(('labels_01', 0), ('labels_02', 1), ('labels_03', 2), etc

		'''
		Plot the force plot for a single example and a single label/output/target
		'''
		print(f'Current label Shown: {list_of_labels[current_label.value]}')

		# print the JS visualization code to the notebook
		shap.initjs()

		shap.force_plot(base_value = explainer.expected_value[current_label.value],
						shap_values = shap_value_single[current_label.value], 
						features = X.iloc[specific_prediction_sample_to_explain,:] # X_idx:X_idx+1
						)

		'''
		Create the summary plot for a specific output/label/target.
		'''
		# Note: We are limiting to the first 50 training examples since it takes time to calculate the full number of sampels
		shap_values = explainer.shap_values(X = X.iloc[0:50,:], nsamples = 100) # X[0:50,:]

		print(f'Current Label Shown: {list_of_labels[current_label.value]}\n')

		# print the JS visualization code to the notebook
		shap.initjs()

		shap.summary_plot(shap_values = shap_values[current_label.value],
				  features = X.iloc[0:50,:],
				  plot_type="bar", # X[0:50,:]
				  show=False
				  )

		plt.savefig('./nowcast_model_summary_plot.png', dpi=150, bbox_inches='tight')

		'''
		Force Plot for the first 50 individual examples.
		'''
		print(f'Current Label Shown: {list_of_labels[current_label.value]}\n')

		# print the JS visualization code to the notebook
		shap.initjs()

		shap.force_plot(base_value = explainer.expected_value[current_label.value],
						shap_values = shap_values[current_label.value], 
						features = X.iloc[0:50,:] # X[0:50,:]
						)


#! /usr/bin/env python

"""
@author: Ajay Arunachalam
Created on: 23/10/2021
Goal: Deep Explainable Forecasting with State-Of-the-Networks for Time-series data - RNN, LSTM, GRU, BiRNN, SNN, GNN, Transformers, GAN, FFNN, etc
Version: 0.0.5
"""

import torch
import torch.nn as nn
import statsmodels.api as sm
from .gnn_layer import *


class Forecasting_Models:


	class RNNModel(nn.Module):

		def __init__(self, input_dim, hidden_dim, layer_dim, output_dim, dropout_prob):
			"""The __init__ method that initiates an RNN instance.

			Args:
				input_dim (int): The number of nodes in the input layer
				hidden_dim (int): The number of nodes in each layer
				layer_dim (int): The number of layers in the network
				output_dim (int): The number of nodes in the output layer
				dropout_prob (float): The probability of nodes being dropped out

			"""
			super(Forecasting_Models.RNNModel, self).__init__()

			# Defining the number of layers and the nodes in each layer
			self.hidden_dim = hidden_dim
			self.layer_dim = layer_dim

			# RNN layers
			self.rnn = nn.RNN(
				input_dim, hidden_dim, layer_dim, batch_first=True, dropout=dropout_prob
			)
			# Fully connected layer
			self.fc = nn.Linear(hidden_dim, output_dim)

		def forward(self, x):
			"""The forward method takes input tensor x and does forward propagation

			Args:
				x (torch.Tensor): The input tensor of the shape (batch size, sequence length, input_dim)

			Returns:
				torch.Tensor: The output tensor of the shape (batch size, output_dim)

			"""
			# Initializing hidden state for first input with zeros
			h0 = torch.zeros(self.layer_dim, x.size(0), self.hidden_dim).requires_grad_()

			# Forward propagation by passing in the input and hidden state into the model
			out, h0 = self.rnn(x, h0.detach())

			# Reshaping the outputs in the shape of (batch_size, seq_length, hidden_size)
			# so that it can fit into the fully connected layer
			out = out[:, -1, :]

			# Convert the final state to our desired output shape (batch_size, output_dim)
			out = self.fc(out)
			return out


	class LSTMModel(nn.Module):
		"""LSTMModel class extends nn.Module class and works as a constructor for LSTMs.

		   LSTMModel class initiates a LSTM module based on PyTorch's nn.Module class.
		   It has only two methods, namely init() and forward(). While the init()
		   method initiates the model with the given input parameters, the forward()
		   method defines how the forward propagation needs to be calculated.
		   Since PyTorch automatically defines back propagation, there is no need
		   to define back propagation method.

		   Attributes:
			   hidden_dim (int): The number of nodes in each layer
			   layer_dim (str): The number of layers in the network
			   lstm (nn.LSTM): The LSTM model constructed with the input parameters.
			   fc (nn.Linear): The fully connected layer to convert the final state
							   of LSTMs to our desired output shape.

		"""
		def __init__(self, input_dim, hidden_dim, layer_dim, output_dim, dropout_prob):
			"""The __init__ method that initiates a LSTM instance.

			Args:
				input_dim (int): The number of nodes in the input layer
				hidden_dim (int): The number of nodes in each layer
				layer_dim (int): The number of layers in the network
				output_dim (int): The number of nodes in the output layer
				dropout_prob (float): The probability of nodes being dropped out

			"""
			super(Forecasting_Models.LSTMModel, self).__init__()

			# Defining the number of layers and the nodes in each layer
			self.hidden_dim = hidden_dim
			self.layer_dim = layer_dim

			# LSTM layers
			self.lstm = nn.LSTM(
				input_dim, hidden_dim, layer_dim, batch_first=True, dropout=dropout_prob
			)

			# Fully connected layer
			self.fc = nn.Linear(hidden_dim, output_dim)

		def forward(self, x):
			"""The forward method takes input tensor x and does forward propagation

			Args:
				x (torch.Tensor): The input tensor of the shape (batch size, sequence length, input_dim)

			Returns:
				torch.Tensor: The output tensor of the shape (batch size, output_dim)

			"""
			# Initializing hidden state for first input with zeros
			h0 = torch.zeros(self.layer_dim, x.size(0), self.hidden_dim).requires_grad_()

			# Initializing cell state for first input with zeros
			c0 = torch.zeros(self.layer_dim, x.size(0), self.hidden_dim).requires_grad_()

			# We need to detach as we are doing truncated backpropagation through time (BPTT)
			# If we don't, we'll backprop all the way to the start even after going through another batch
			# Forward propagation by passing in the input, hidden state, and cell state into the model
			out, (hn, cn) = self.lstm(x, (h0.detach(), c0.detach()))

			# Reshaping the outputs in the shape of (batch_size, seq_length, hidden_size)
			# so that it can fit into the fully connected layer
			out = out[:, -1, :]

			# Convert the final state to our desired output shape (batch_size, output_dim)
			out = self.fc(out)

			return out

	class GRUModel(nn.Module):
		"""GRUModel class extends nn.Module class and works as a constructor for GRUs.

		   GRUModel class initiates a GRU module based on PyTorch's nn.Module class.
		   It has only two methods, namely init() and forward(). While the init()
		   method initiates the model with the given input parameters, the forward()
		   method defines how the forward propagation needs to be calculated.
		   Since PyTorch automatically defines back propagation, there is no need
		   to define back propagation method.

		   Attributes:
			   hidden_dim (int): The number of nodes in each layer
			   layer_dim (str): The number of layers in the network
			   gru (nn.GRU): The GRU model constructed with the input parameters.
			   fc (nn.Linear): The fully connected layer to convert the final state
							   of GRUs to our desired output shape.

		"""
		def __init__(self, input_dim, hidden_dim, layer_dim, output_dim, dropout_prob):
			"""The __init__ method that initiates a GRU instance.

			Args:
				input_dim (int): The number of nodes in the input layer
				hidden_dim (int): The number of nodes in each layer
				layer_dim (int): The number of layers in the network
				output_dim (int): The number of nodes in the output layer
				dropout_prob (float): The probability of nodes being dropped out

			"""
			super(Forecasting_Models.GRUModel, self).__init__()

			# Defining the number of layers and the nodes in each layer
			self.layer_dim = layer_dim
			self.hidden_dim = hidden_dim

			# GRU layers
			self.gru = nn.GRU(
				input_dim, hidden_dim, layer_dim, batch_first=True, dropout=dropout_prob
			)

			# Fully connected layer
			self.fc = nn.Linear(hidden_dim, output_dim)

		def forward(self, x):
			"""The forward method takes input tensor x and does forward propagation

			Args:
				x (torch.Tensor): The input tensor of the shape (batch size, sequence length, input_dim)

			Returns:
				torch.Tensor: The output tensor of the shape (batch size, output_dim)

			"""
			# Initializing hidden state for first input with zeros
			h0 = torch.zeros(self.layer_dim, x.size(0), self.hidden_dim).requires_grad_()

			# Forward propagation by passing in the input and hidden state into the model
			out, _ = self.gru(x, h0.detach())

			# Reshaping the outputs in the shape of (batch_size, seq_length, hidden_size)
			# so that it can fit into the fully connected layer
			out = out[:, -1, :]

			# Convert the final state to our desired output shape (batch_size, output_dim)
			out = self.fc(out)

			return out

	
	# Bidirectional recurrent neural network (many-to-one)
	class BiRNNModel(nn.Module):

		def __init__(self, input_dim, hidden_dim, layer_dim, output_dim, dropout_prob):
			super(Forecasting_Models.BiRNNModel, self).__init__()
			self.hidden_dim = hidden_dim
			self.layer_dim = layer_dim
			self.lstm = nn.LSTM(input_size, hidden_dim, layer_dim, batch_first=True, bidirectional=True, dropout=dropout_prob)
			self.fc = nn.Linear(hidden_dim*2, output_dim)  # 2 for bidirection
		
		def forward(self, x):
			# Set initial states
			h0 = torch.zeros(self.layer_dim*2, x.size(0), self.hidden_dim).requires_grad_() #.to(device) # 2 for bidirection 
			c0 = torch.zeros(self.layer_dim*2, x.size(0), self.hidden_dim).requires_grad_() #.to(device)
			
			# Forward propagate LSTM
			out, _ = self.lstm(x, (h0.detach(), c0.detach()))  # out: tensor of shape (batch_size, seq_length, hidden_size*2)
			
			# Decode the hidden state of the last time step
			out = self.fc(out[:, -1, :])
			return out

	class BiGRUModel(nn.Module):
		"""BiGRUModel class extends nn.Module class and works as a constructor for GRUs.

		   BiGRUModel class initiates a GRU module based on PyTorch's nn.Module class.
		   It has only two methods, namely init() and forward(). While the init()
		   method initiates the model with the given input parameters, the forward()
		   method defines how the forward propagation needs to be calculated.
		   Since PyTorch automatically defines back propagation, there is no need
		   to define back propagation method.

		   Attributes:
			   hidden_dim (int): The number of nodes in each layer
			   layer_dim (str): The number of layers in the network
			   gru (nn.GRU): The GRU model constructed with the input parameters.
			   fc (nn.Linear): The fully connected layer to convert the final state
							   of GRUs to our desired output shape.

		"""
		def __init__(self, input_dim, hidden_dim, layer_dim, output_dim, dropout_prob):
			"""The __init__ method that initiates a GRU instance.

			Args:
				input_dim (int): The number of nodes in the input layer
				hidden_dim (int): The number of nodes in each layer
				layer_dim (int): The number of layers in the network
				output_dim (int): The number of nodes in the output layer
				dropout_prob (float): The probability of nodes being dropped out

			"""
			super(Forecasting_Models.BiGRUModel, self).__init__()

			# Defining the number of layers and the nodes in each layer
			self.layer_dim = layer_dim
			self.hidden_dim = hidden_dim

			# GRU layers
			self.gru = nn.GRU(
				input_dim, hidden_dim, layer_dim, batch_first=True, bidirectional=True, dropout=dropout_prob
			)

			# Fully connected layer
			self.fc = nn.Linear(hidden_dim*2, output_dim) # 2 for bidirection

		def forward(self, x):
			"""The forward method takes input tensor x and does forward propagation

			Args:
				x (torch.Tensor): The input tensor of the shape (batch size, sequence length, input_dim)

			Returns:
				torch.Tensor: The output tensor of the shape (batch size, output_dim)

			"""
			# Initializing hidden state for first input with zeros
			h0 = torch.zeros(self.layer_dim*2, x.size(0), self.hidden_dim).requires_grad_() # 2 for bidirection

			# Forward propagation by passing in the input and hidden state into the model
			out, _ = self.gru(x, h0.detach())

			# Reshaping the outputs in the shape of (batch_size, seq_length, hidden_size)
			# so that it can fit into the fully connected layer
			out = out[:, -1, :]

			# Convert the final state to our desired output shape (batch_size, output_dim)
			out = self.fc(out)

			return out

	class SNNModel(nn.Module):
		pass


	class CNNModel(nn.Module):

		def __init__(self, input_dim, hidden_dim, layer_dim, output_dim, dropout_prob):
			"""The __init__ method that initiates an CNN instance.

			Args:
				input_dim (int): The number of nodes in the input layer
				hidden_dim (int): The number of nodes in each layer
				layer_dim (int): The number of layers in the network
				output_dim (int): The number of nodes in the output layer
				dropout_prob (float): The probability of nodes being dropped out

			"""
			super(Forecasting_Models.CNNModel, self).__init__()

			# Defining the number of layers and the nodes in each layer
			self.hidden_dim = hidden_dim
			self.layer_dim = layer_dim

			# CNN layers
			self.cnn = nn.CNN(
				input_dim, hidden_dim, layer_dim, batch_first=True, dropout=dropout_prob
			)
			# Fully connected layer
			self.fc = nn.Linear(hidden_dim, output_dim)

		def forward(self, x):
			"""The forward method takes input tensor x and does forward propagation

			Args:
				x (torch.Tensor): The input tensor of the shape (batch size, sequence length, input_dim)

			Returns:
				torch.Tensor: The output tensor of the shape (batch size, output_dim)

			"""
			# Initializing hidden state for first input with zeros
			h0 = torch.zeros(self.layer_dim, x.size(0), self.hidden_dim).requires_grad_()

			# Forward propagation by passing in the input and hidden state into the model
			out, h0 = self.cnn(x, h0.detach())

			# Reshaping the outputs in the shape of (batch_size, seq_length, hidden_size)
			# so that it can fit into the fully connected layer
			out = out[:, -1, :]

			# Convert the final state to our desired output shape (batch_size, output_dim)
			out = self.fc(out)
			return out


	class GNNModel(nn.Module):

		def __init__(self, input_dim, hidden_dim, layer_dim, output_dim, dropout_prob):
			
		 	# self, in_features = 7, hidden_dim = 64, classes = 2, dropout = 0.5

			super(Forecasting_Models.GNNModel, self).__init__()

			# GNN layers
			self.conv1 = GCN_Layer(input_dim, hidden_dim) # in_features, hidden_dim
			self.conv2 = GCN_Layer(hidden_dim, hidden_dim)
			self.conv3 = GCN_Layer(hidden_dim, hidden_dim)
			self.fc = nn.Linear(hidden_dim, output_dim)
			self.dropout = dropout_prob

		def forward(self, x, A):
			x = self.conv1(x, A)
			x = F.relu(x)
			x = self.conv2(x, A)
			x = F.relu(x)
			x = self.conv3(x, A)
			x = F.dropout(x, p=self.dropout, training=self.training)
			# aggregate node embeddings
			x = x.mean(dim=1)
			# final layer
			return self.fc(x)

	# Deep CNN

	class DeepCNNModel(nn.Module):

		"""
		Model : Class for DeepCNN model
		"""
		def __init__(self, input_dim, hidden_dim, layer_dim, output_dim, dropout_prob): # self, LOOKBACK_SIZE, DIMENSION, KERNEL_SIZE
			super(Forecasting_Models.DeepCNNModel, self).__init__()
			import torch
			self.conv1d_1_layer = torch.nn.Conv1d(in_channels=2, out_channels=16, kernel_size=3) # , stride=2
			self.relu_1_layer = torch.nn.ReLU()
			self.maxpooling_1_layer = torch.nn.MaxPool1d(kernel_size=(3)-1)
			self.conv1d_2_layer = torch.nn.Conv1d(in_channels=16, out_channels=16, kernel_size=3)
			self.relu_2_layer = torch.nn.ReLU()
			self.maxpooling_2_layer = torch.nn.MaxPool1d(kernel_size=(3)-1)
			self.flatten_layer = torch.nn.Flatten()
			self.dense_1_layer = torch.nn.Linear(80, 40)
			self.relu_3_layer = torch.nn.ReLU()
			self.dropout_layer = torch.nn.Dropout(p=dropout_prob)
			self.dense_2_layer = torch.nn.Linear(40, output_dim)
			
		def forward(self, x):
			x = self.conv1d_1_layer(x)
			x = self.relu_1_layer(x)
			x = self.maxpooling_1_layer(x)
			x = self.conv1d_2_layer(x)
			x = self.relu_2_layer(x)
			x = self.maxpooling_2_layer(x)
			x = self.flatten_layer(x)
			x = self.dense_1_layer(x)
			x = self.relu_3_layer(x)
			x = self.dropout_layer(x)
			return self.dense_2_layer(x)

	# GCN

	'''
	import torch
	from torch import nn

	class GCN(nn.Module):

		def __init__(self, *sizes):
			super().__init__()
			self.layers = nn.ModuleList([
				nn.Linear(x, y) for x, y in zip(sizes[:-1], sizes[1:])
			])
		def forward(self, vertices, edges):
			# ----- Build the adjacency matrix -----
			# Start with self-connections
			adj = torch.eye(len(vertices))
			# edges contain connected vertices: [vertex_0, vertex_1] 
			adj[edges[:, 0], edges[:, 1]] = 1 
			adj[edges[:, 1], edges[:, 0]] = 1
			
			# ----- Forward data pass -----
			for layer in self.layers:
				vertices = torch.sigmoid(layer(adj @ vertices))
	 
			return vertices
	'''

	'''
	# FFNN

	class FFModel(torch.nn.Module):

		def __init__(self, DIMENSION):
			super(Anamoly.MPLR, self).__init__()
			import torch
			self.layer_1 = torch.nn.Linear(DIMENSION, 16)
			self.layer_2 = torch.nn.Linear(16, 32)
			self.layer_3 = torch.nn.Linear(32, 16)
			self.layer_out = torch.nn.Linear(16, 1)
			self.relu = torch.nn.ReLU()

		def forward(self, inputs):
			x = self.relu(self.layer_1(inputs))
			x = self.relu(self.layer_2(x))
			x = self.relu(self.layer_3(x))
			x = self.layer_out(x)
			return (x)

		def predict(self, test_inputs):
			x = self.relu(self.layer_1(test_inputs))
			x = self.relu(self.layer_2(x))
			x = self.relu(self.layer_3(x))
			x = self.layer_out(x)
			return (x)

	# Deep CNN

	class DeepCNNModel(torch.nn.Module):
		"""
		Model : Class for DeepCNN model
		"""
		def __init__(self, LOOKBACK_SIZE, DIMENSION, KERNEL_SIZE):
			super(Anamoly.DeepCNN, self).__init__()
			import torch
			self.conv1d_1_layer = torch.nn.Conv1d(in_channels=LOOKBACK_SIZE, out_channels=16, kernel_size=KERNEL_SIZE) # , stride=2
			self.relu_1_layer = torch.nn.ReLU()
			self.maxpooling_1_layer = torch.nn.MaxPool1d(kernel_size=(KERNEL_SIZE)-1)
			self.conv1d_2_layer = torch.nn.Conv1d(in_channels=16, out_channels=16, kernel_size=KERNEL_SIZE)
			self.relu_2_layer = torch.nn.ReLU()
			self.maxpooling_2_layer = torch.nn.MaxPool1d(kernel_size=(KERNEL_SIZE)-1)
			self.flatten_layer = torch.nn.Flatten()
			self.dense_1_layer = torch.nn.Linear(80, 40)
			self.relu_3_layer = torch.nn.ReLU()
			self.dropout_layer = torch.nn.Dropout(p=0.25)
			self.dense_2_layer = torch.nn.Linear(40, DIMENSION)
			
		def forward(self, x):
			x = self.conv1d_1_layer(x)
			x = self.relu_1_layer(x)
			x = self.maxpooling_1_layer(x)
			x = self.conv1d_2_layer(x)
			x = self.relu_2_layer(x)
			x = self.maxpooling_2_layer(x)
			x = self.flatten_layer(x)
			x = self.dense_1_layer(x)
			x = self.relu_3_layer(x)
			x = self.dropout_layer(x)
			return self.dense_2_layer(x)

	# LSTMAENN

	class LSTMAEModel(torch.nn.Module):
		"""
		Model : Class for LSTMAENN model
		"""
		def __init__(self, LOOKBACK_SIZE, DIMENSION):
			import torch
			super(Anamoly.LSTMAENN, self).__init__()
			self.lstm_1_layer = torch.nn.LSTM(DIMENSION, 128, 1)
			self.dropout_1_layer = torch.nn.Dropout(p=0.2)
			self.lstm_2_layer = torch.nn.LSTM(128, 64, 1)
			self.dropout_2_layer = torch.nn.Dropout(p=0.2)
			self.lstm_3_layer = torch.nn.LSTM(64, 64, 1)
			self.dropout_3_layer = torch.nn.Dropout(p=0.2)
			self.lstm_4_layer = torch.nn.LSTM(64, 128, 1)
			self.dropout_4_layer = torch.nn.Dropout(p=0.2)
			self.linear_layer = torch.nn.Linear(128, DIMENSION)
			
		def forward(self, x):
			x, (_,_) = self.lstm_1_layer(x)
			x = self.dropout_1_layer(x)
			x, (_,_) = self.lstm_2_layer(x)
			x = self.dropout_2_layer(x)
			x, (_,_) = self.lstm_3_layer(x)
			x = self.dropout_3_layer(x)
			x, (_,_) = self.lstm_4_layer(x)
			x = self.dropout_4_layer(x)
			return self.linear_layer(x)

	# GNN

	# GAN

	class GAN(torch.nn.Module):
		"""
		Model: Class for Generative Adversarial Networks
		"""
		pass

	# TRANSFORMER

	class TF(torch.nn.Module):
		"""
		Model: Class for Transformer Networks
		"""
		pass


	# SNN

	# https://github.com/ajayarunachalam/pynmsnn/blob/main/pyNM/spiking_regressor.py

	class SNN(torch.nn.Module):
		"""
		Model: Class for Spiking Neural Networks
		"""
		pass


	# Using the auto time-series functionality 
	class AutoTimeSeries:
		from auto_ts import auto_timeseries

		def __init__(self, df, ts_column, time_interval, seasonality, seasonal_period, target, FORECAST_PERIOD):
			super(Anamoly.AutoTimeSeries, self).__init__()

		
		# Train Model

		self.ts_column =  ts_column
		self.target = target
		self.FORECAST_PERIOD = FORECAST_PERIOD
		self.interval = interval
		sep = ','

		#FORECAST_PERIOD = 2

		model = auto_timeseries(score_type='rmse',
						time_interval='QS',
						non_seasonal_pdq=None, seasonality=False, seasonal_period=12,
						model_type=['prophet'],  # model_type available = ['best', 'prophet', 'stats', 'ml', 'arima','ARIMA','Prophet','SARIMAX', 'VAR', 'ML']
						verbose=2)

		model.fit(
			traindata=dataset,
			# traindata=file_path,  # Alternately, you can specify the file directly
			ts_column=ts_column,
			target=target,
			cv=5,
			sep=sep)

		model.get_leaderboard()

		# Make sure all models have the same number of CV folds
		model.get_cv_scores()

		model.plot_cv_scores()

		# Using Machine Learning Model
		future_predictions = model.predict(
			testdata=FORECAST_PERIOD
		)  

		# Forecasted Future Predictions dataframe
		future_predictions

	'''


	


#! /usr/bin/env python


# Copyright (C) 2021-2022 Ajay Arunachalam <ajay.arunachalam08@gmail.com>
# License: MIT, ajay.arunachalam08@gmail.com

import pip
import logging
import pkg_resources
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


__version__ = '0.0.5'

def readme():
    with open('README.rst', 'r', encoding='utf-8') as f:
        return f.read()


setup(
    name='deep_xf',
    version=__version__,
    packages=["deep_xf"],
    description='DEEPXF - An open-source, low-code explainable forecasting and nowcasting library with state-of-the-art deep neural networks and Dynamic Factor Model. Now available with additional addons like Denoising TS signals with ensembling of filters, TS signal similarity test with Siamese Neural Networks',
    long_description = readme(),
    long_description_content_type="text/markdown",
    url='https://github.com/ajayarunachalam/Deep_XF',
    install_requires=[
        "ipython",
        "jupyter",
        "tqdm",
        "pandas",
        "matplotlib",
        "seaborn==0.9.0",
        "scikit-learn>=0.24.0",
        "pandas_profiling",
        "statsmodels==0.12.2",
        "keras",
        "torch",
        "shap==0.39.0",
        "py-ecg-detectors",
    ],
    license='MIT',
    include_package_data=True,
    author='Ajay Arunachalam',
    author_email='ajay.arunachalam08@gmail.com')
